package net.md_5.bungee;

import net.md_5.bungee.scheduler.*;
import io.github.waterfallmc.waterfall.conf.*;
import net.md_5.bungee.util.*;
import java.util.concurrent.locks.*;
import net.md_5.bungee.conf.*;
import com.google.gson.*;
import java.lang.reflect.*;
import net.md_5.bungee.api.chat.*;
import net.md_5.bungee.chat.*;
import io.github.waterfallmc.waterfall.log4j.*;
import net.md_5.bungee.compress.*;
import edu.umd.cs.findbugs.annotations.*;
import net.md_5.bungee.api.plugin.*;
import com.google.common.util.concurrent.*;
import net.md_5.bungee.netty.*;
import java.util.concurrent.*;
import io.netty.bootstrap.*;
import io.netty.util.*;
import io.netty.util.concurrent.*;
import net.md_5.bungee.query.*;
import io.netty.channel.*;
import com.arkflame.flamecord.config.*;
import java.util.logging.*;
import com.arkflame.flamecord.firewall.*;
import net.md_5.bungee.api.config.*;
import java.net.*;
import java.io.*;
import java.text.*;
import net.md_5.bungee.api.connection.*;
import java.util.*;
import net.md_5.bungee.protocol.*;
import net.md_5.bungee.command.*;
import net.md_5.bungee.protocol.packet.*;
import com.google.common.base.*;
import com.google.common.collect.*;
import net.md_5.bungee.api.*;
import net.md_5.bungee.api.scheduler.*;

//New imports
import net.md_5.bungee.api.Title;
import java.util.Timer;
import java.util.TimerTask;

public class BungeeCord extends ProxyServer
{
    private String version;
    public volatile boolean isRunning;
    public final Configuration config;
    private ResourceBundle baseBundle;
    private ResourceBundle customBundle;
    public EventLoopGroup bossEventLoopGroup;
    public EventLoopGroup workerEventLoopGroup;
    private final Timer saveThread;
    private final Collection<Channel> listeners;
    private final Map<String, UserConnection> connections;
    private final Map<UUID, UserConnection> connectionsByOfflineUUID;
    private final Map<UUID, UserConnection> connectionsByUUID;
    private final ReadWriteLock connectionLock;
    public final PluginManager pluginManager;
    private ReconnectHandler reconnectHandler;
    private ConfigurationAdapter configurationAdapter;
    private final Collection<String> pluginChannels;
    private final File pluginsFolder;
    private final BungeeScheduler scheduler;
    private final Logger logger;
    public final Gson gson;
    public final Gson gsonLegacy;
    private ConnectionThrottle connectionThrottle;
    private Flamecord flamecord;

    public static BungeeCord getInstance() {
        return (BungeeCord)ProxyServer.getInstance();
    }

    @SuppressFBWarnings({ "DM_DEFAULT_ENCODING" })
    public BungeeCord() throws IOException {
        this.version = "1.3-Final-SPARK";
        this.config = (Configuration)new WaterfallConfiguration();
        this.saveThread = new Timer("Reconnect Saver");
        this.listeners = new HashSet<Channel>();
        this.connections = (Map<String, UserConnection>)new CaseInsensitiveMap();
        this.connectionsByOfflineUUID = new HashMap<UUID, UserConnection>();
        this.connectionsByUUID = new HashMap<UUID, UserConnection>();
        this.connectionLock = new ReentrantReadWriteLock();
        this.configurationAdapter = new YamlConfig();
        this.pluginChannels = new HashSet<String>();
        this.pluginsFolder = new File("plugins");
        this.scheduler = new BungeeScheduler();
        this.gson = new GsonBuilder().registerTypeAdapter((Type)BaseComponent.class, (Object)new ComponentSerializer()).registerTypeAdapter((Type)TextComponent.class, (Object)new TextComponentSerializer()).registerTypeAdapter((Type)TranslatableComponent.class, (Object)new TranslatableComponentSerializer()).registerTypeAdapter((Type)KeybindComponent.class, (Object)new KeybindComponentSerializer()).registerTypeAdapter((Type)ScoreComponent.class, (Object)new ScoreComponentSerializer()).registerTypeAdapter((Type)SelectorComponent.class, (Object)new SelectorComponentSerializer()).registerTypeAdapter((Type)ServerPing.PlayerInfo.class, (Object)new PlayerInfoSerializer()).registerTypeAdapter((Type)Favicon.class, (Object)Favicon.getFaviconTypeAdapter()).create();
        this.gsonLegacy = new GsonBuilder().registerTypeAdapter((Type)BaseComponent.class, (Object)new ComponentSerializer()).registerTypeAdapter((Type)TextComponent.class, (Object)new TextComponentSerializer()).registerTypeAdapter((Type)TranslatableComponent.class, (Object)new TranslatableComponentSerializer()).registerTypeAdapter((Type)ServerPing.PlayerInfo.class, (Object)new PlayerInfoSerializer(4)).registerTypeAdapter((Type)Favicon.class, (Object)Favicon.getFaviconTypeAdapter()).create();
        this.registerChannel("BungeeCord");
        Preconditions.checkState(new File(".").getAbsolutePath().indexOf(33) == -1, (Object)"Cannot use Flamecord in directory with ! in path.");
        try {
            this.baseBundle = ResourceBundle.getBundle("messages");
        }
        catch (MissingResourceException ex) {
            this.baseBundle = ResourceBundle.getBundle("messages", Locale.ENGLISH);
        }
        this.reloadMessages();
        System.setProperty("library.jansi.version", "BungeeCord");
        this.logger = WaterfallLogger.create();
        this.pluginManager = new PluginManager(this);
        if (!Boolean.getBoolean("net.md_5.bungee.native.disable")) {
            if (EncryptionUtil.nativeFactory.load()) {
                this.logger.info("Using mbed TLS based native cipher.");
            }
            else {
                this.logger.info("Using standard Java JCE cipher.");
            }
            if (CompressFactory.zlib.load()) {
                this.logger.info("Using zlib based native compressor.");
            }
            else {
                this.logger.info("Using standard Java compressor.");
            }
        }
    }

    private void registerModules() {
        if (ModulesConfig.reloadEnabled) {
            this.getPluginManager().registerCommand(null, new CommandReload());
        }
        if (ModulesConfig.ipEnabled) {
            this.getPluginManager().registerCommand(null, new CommandIP());
        }
        if (ModulesConfig.permsEnabled) {
            this.getPluginManager().registerCommand(null, new CommandPerms());
        }
        this.getPluginManager().registerCommand(null, new CommandEnd());
        this.getPluginManager().registerCommand(null, new CommandBungee());
        if (ModulesConfig.alertEnabled) {
            this.getPluginManager().registerCommand(null, new CommandAlert());
        }
        if (ModulesConfig.findEnabled) {
            this.getPluginManager().registerCommand(null, new CommandFind());
        }
        if (ModulesConfig.listEnabled) {
            this.getPluginManager().registerCommand(null, new CommandList());
        }
        if (ModulesConfig.sendEnabled) {
            this.getPluginManager().registerCommand(null, new CommandSend());
        }
        if (ModulesConfig.serverEnabled) {
            this.getPluginManager().registerCommand(null, new CommandServer());
        }
        if (ModulesConfig.reconnectEnabled) {
            for (final ListenerInfo info : this.getConfig().getListeners()) {
                if (!info.isForceDefault() && this.getReconnectHandler() == null) {
                    this.setReconnectHandler(new YamlReconnectHandler());
                    break;
                }
            }
        }
        this.getPluginManager().registerCommand(null, new CommandFlamecord());
        this.getPluginManager().registerCommand(null, new CommandFirewall());
    }

    @SuppressFBWarnings({ "RV_RETURN_VALUE_IGNORED_BAD_PRACTICE" })
    public void start() throws Exception {
        System.setProperty("io.netty.selectorAutoRebuildThreshold", "0");
        if (System.getProperty("io.netty.leakDetectionLevel") == null) {
            ResourceLeakDetector.setLevel(ResourceLeakDetector.Level.DISABLED);
        }
        this.bossEventLoopGroup = PipelineUtils.newEventLoopGroup(0, new ThreadFactoryBuilder().setNameFormat("Netty Boss IO Thread #%1$d").build());
        this.workerEventLoopGroup = PipelineUtils.newEventLoopGroup(0, new ThreadFactoryBuilder().setNameFormat("Netty Worker IO Thread #%1$d").build());
        this.pluginsFolder.mkdir();
        this.pluginManager.detectPlugins(this.pluginsFolder);
        this.pluginManager.loadPlugins();
        this.config.load();
        this.registerModules();
        this.flamecord = new Flamecord();
        if (this.config.isForgeSupport()) {
            this.registerChannel("FML");
            this.registerChannel("FML|HS");
            this.registerChannel("FORGE");
        }
        this.isRunning = true;
        this.pluginManager.enablePlugins();
        if (this.config.getThrottle() > 0) {
            this.connectionThrottle = new ConnectionThrottle(this.config.getThrottle(), this.config.getThrottleLimit());
        }
        this.startListeners();
        this.saveThread.scheduleAtFixedRate(new TimerTask() {
            @Override
            public void run() {
                if (BungeeCord.this.getReconnectHandler() != null) {
                    BungeeCord.this.getReconnectHandler().save();
                }
            }
        }, 0L, TimeUnit.MINUTES.toMillis(5L));
        this.reloadProxy();
    }

    public void startListeners() {
        for (final ListenerInfo info : this.config.getListeners()) {
            if (info.isProxyProtocol()) {
                this.getLogger().log(Level.WARNING, "Using PROXY protocol for listener {0}, please ensure this listener is adequately firewalled.", info.getHost());
                if (this.connectionThrottle != null) {
                    this.connectionThrottle = null;
                    this.getLogger().log(Level.WARNING, "Since PROXY protocol is in use, internal connection throttle has been disabled.");
                }
            }
            final ChannelFutureListener listener = (ChannelFutureListener)new ChannelFutureListener() {
                public void operationComplete(final ChannelFuture future) throws Exception {
                    if (future.isSuccess()) {
                        BungeeCord.this.listeners.add(future.channel());
                        BungeeCord.this.getLogger().log(Level.INFO, "Listening on {0}", info.getHost());
                    }
                    else {
                        BungeeCord.this.getLogger().log(Level.WARNING, "Could not bind to host " + info.getHost(), future.cause());
                    }
                }
            };
            ((ServerBootstrap)((ServerBootstrap)(ServerBootstrap)new ServerBootstrap().channel(PipelineUtils.getServerChannel())).option(ChannelOption.SO_REUSEADDR, true).childAttr(PipelineUtils.LISTENER, info).childHandler(PipelineUtils.SERVER_CHILD).group(this.bossEventLoopGroup, this.workerEventLoopGroup).localAddress(info.getHost())).bind().addListener(listener);
            if (info.isQueryEnabled()) {
                final ChannelFutureListener bindListener = (ChannelFutureListener)new ChannelFutureListener() {
                    public void operationComplete(final ChannelFuture future) throws Exception {
                        if (future.isSuccess()) {
                            BungeeCord.this.listeners.add(future.channel());
                            BungeeCord.this.getLogger().log(Level.INFO, "Started query on {0}", future.channel().localAddress());
                        }
                        else {
                            BungeeCord.this.getLogger().log(Level.WARNING, "Could not bind to host " + info.getHost(), future.cause());
                        }
                    }
                };
                new RemoteQuery(this, info).start((Class<? extends Channel>)PipelineUtils.getDatagramChannel(), new InetSocketAddress(info.getHost().getAddress(), info.getQueryPort()), this.workerEventLoopGroup, bindListener);
            }
        }
    }

    public void stopListeners() {
        for (final Channel listener : this.listeners) {
            this.getLogger().log(Level.INFO, "Closing listener {0}", listener);
            try {
                listener.close().syncUninterruptibly();
            }
            catch (ChannelException ex) {
                this.getLogger().severe("Could not close listen thread");
            }
        }
        this.listeners.clear();
    }

    @Override
    public void stop() {
        this.stop(ChatColor.translateAlternateColorCodes('&', MessagesConfig.restart));
    }

    @Override
    public synchronized void stop(final String reason) {
        if (!this.isRunning) {
            return;
        }
        this.isRunning = false;
        new Thread("Shutdown Thread") {
            @SuppressFBWarnings({ "DM_EXIT" })
            @Override
            public void run() {
                BungeeCord.this.stopListeners();
                BungeeCord.this.getLogger().info("Closing pending connections");
                BungeeCord.this.connectionLock.readLock().lock();
                try {
                    BungeeCord.this.getLogger().log(Level.INFO, "Disconnecting {0} connections", BungeeCord.this.connections.size());
                    for (final UserConnection user : BungeeCord.this.connections.values()) {
                        user.disconnect(reason);
                    }
                }
                finally {
                    BungeeCord.this.connectionLock.readLock().unlock();
                }
                try {
                    Thread.sleep(500L);
                }
                catch (InterruptedException ex) {}
                if (BungeeCord.this.reconnectHandler != null) {
                    BungeeCord.this.getLogger().info("Saving reconnect locations");
                    BungeeCord.this.reconnectHandler.save();
                    BungeeCord.this.reconnectHandler.close();
                }
                BungeeCord.this.saveThread.cancel();
                BungeeCord.this.getLogger().info("Disabling plugins");
                for (Plugin plugin : Lists.reverse( new ArrayList<>( pluginManager.getPlugins()))) {
                    try {
                        plugin.onDisable();
                        for (final Handler handler : plugin.getLogger().getHandlers()) {
                            handler.close();
                        }
                    }
                    catch (Throwable t) {
                        BungeeCord.this.getLogger().log(Level.SEVERE, "Exception disabling plugin " + plugin.getDescription().getName(), t);
                    }
                    BungeeCord.this.getScheduler().cancel(plugin);
                    plugin.getExecutorService().shutdownNow();
                }
                BungeeCord.this.getLogger().info("Closing IO threads");
                BungeeCord.this.bossEventLoopGroup.shutdownGracefully();
                BungeeCord.this.workerEventLoopGroup.shutdownGracefully();
                while (true) {
                    try {
                        BungeeCord.this.bossEventLoopGroup.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
                        BungeeCord.this.workerEventLoopGroup.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
                    }
                    catch (InterruptedException ex2) {
                        continue;
                    }
                    break;
                }
                BungeeCord.this.getLogger().info("Thank you and goodbye");
                for (final Handler handler2 : BungeeCord.this.getLogger().getHandlers()) {
                    handler2.close();
                }
                System.exit(0);
            }
        }.start();
    }

    public void broadcast(final DefinedPacket packet) {
        this.connectionLock.readLock().lock();
        try {
            for (final UserConnection con : this.connections.values()) {
                con.unsafe().sendPacket(packet);
            }
        }
        finally {
            this.connectionLock.readLock().unlock();
        }
    }

    @Override
    public String getName() {
        return "FlameCord";
    }

    @Override
    public String getVersion() {
        return this.version;
    }

    public void reloadProxy() {
        Firewall.serverAdresses.clear();
        for (final ServerInfo server : this.getConfig().getServersCopy().values()) {
            final InetAddress address = server.getAddress().getAddress();
            Firewall.serverAdresses.add(address);
        }
    }

    public void reloadMessages() {
        final File file = new File("messages.properties");
        if (file.isFile()) {
            try (final FileReader rd = new FileReader(file)) {
                this.customBundle = new PropertyResourceBundle(rd);
            }
            catch (IOException ex) {
                this.getLogger().log(Level.SEVERE, "Could not load custom messages.properties", ex);
            }
        }
    }

    @Override
    public String getTranslation(final String name, final Object... args) {
        String translation = "<translation '" + name + "' missing>";
        try {
            final String string = (this.customBundle != null && this.customBundle.containsKey(name)) ? this.customBundle.getString(name) : this.baseBundle.getString(name);
            translation = ((args.length == 0) ? string : MessageFormat.format(string, args));
        }
        catch (MissingResourceException ex) {}
        return translation;
    }

    @Override
    public Collection<ProxiedPlayer> getPlayers() {
        this.connectionLock.readLock().lock();
        try {
            return Collections.unmodifiableCollection((Collection<? extends ProxiedPlayer>)new HashSet<ProxiedPlayer>(this.connections.values()));
        }
        finally {
            this.connectionLock.readLock().unlock();
        }
    }

    @Override
    public int getOnlineCount() {
        return this.connections.size();
    }

    @Override
    public ProxiedPlayer getPlayer(final String name) {
        this.connectionLock.readLock().lock();
        try {
            return this.connections.get(name);
        }
        finally {
            this.connectionLock.readLock().unlock();
        }
    }

    public UserConnection getPlayerByOfflineUUID(final UUID name) {
        this.connectionLock.readLock().lock();
        try {
            return this.connectionsByOfflineUUID.get(name);
        }
        finally {
            this.connectionLock.readLock().unlock();
        }
    }

    @Override
    public ProxiedPlayer getPlayer(final UUID uuid) {
        this.connectionLock.readLock().lock();
        try {
            return this.connectionsByUUID.get(uuid);
        }
        finally {
            this.connectionLock.readLock().unlock();
        }
    }

    @Override
    public Map<String, ServerInfo> getServers() {
        return (Map<String, ServerInfo>)this.config.getServers();
    }

    @Override
    public Map<String, ServerInfo> getServersCopy() {
        return this.config.getServersCopy();
    }

    @Override
    public ServerInfo getServerInfo(final String name) {
        return this.config.getServerInfo(name);
    }

    @Override
    public void registerChannel(final String channel) {
        synchronized (this.pluginChannels) {
            this.pluginChannels.add(channel);
        }
    }

    @Override
    public void unregisterChannel(final String channel) {
        synchronized (this.pluginChannels) {
            this.pluginChannels.remove(channel);
        }
    }

    @Override
    public Collection<String> getChannels() {
        synchronized (this.pluginChannels) {
            return Collections.unmodifiableCollection((Collection<? extends String>)this.pluginChannels);
        }
    }

    public PluginMessage registerChannels(final int protocolVersion) {
        if (protocolVersion >= 393) {
            return new PluginMessage("minecraft:register", Util.format(Iterables.transform((Iterable)this.pluginChannels, (Function)PluginMessage.MODERNISE), "\u0000").getBytes(Charsets.UTF_8), false);
        }
        return new PluginMessage("REGISTER", Util.format(this.pluginChannels, "\u0000").getBytes(Charsets.UTF_8), false);
    }

    @Override
    public int getProtocolVersion() {
        return ProtocolConstants.SUPPORTED_VERSION_IDS.get(ProtocolConstants.SUPPORTED_VERSION_IDS.size() - 1);
    }

    @Override
    public String getGameVersion() {
        return this.getConfig().getGameVersion();
    }

    @Override
    public ServerInfo constructServerInfo(final String name, final InetSocketAddress address, final String motd, final boolean restricted) {
        return new BungeeServerInfo(name, address, motd, restricted);
    }

    @Override
    public CommandSender getConsole() {
        return ConsoleCommandSender.getInstance();
    }

    @Override
    public void broadcast(final String message) {
        this.broadcast(TextComponent.fromLegacyText(message));
    }

    @Override
    public void broadcast(final BaseComponent... message) {
        this.getConsole().sendMessage(BaseComponent.toLegacyText(message));
        this.broadcast((DefinedPacket)new Chat(ComponentSerializer.toString(message)));
    }

    @Override
    public void broadcast(final BaseComponent message) {
        this.getConsole().sendMessage(message.toLegacyText());
        this.broadcast((DefinedPacket)new Chat(ComponentSerializer.toString(message)));
    }

    public void addConnection(final UserConnection con) {
        this.connectionLock.writeLock().lock();
        try {
            this.connections.put(con.getName(), con);
            this.connectionsByUUID.put(con.getUniqueId(), con);
            this.connectionsByOfflineUUID.put(con.getPendingConnection().getOfflineId(), con);
        }
        finally {
            this.connectionLock.writeLock().unlock();
        }
    }

    public void removeConnection(final UserConnection con) {
        this.connectionLock.writeLock().lock();
        try {
            if (this.connections.get(con.getName()) == con) {
                this.connections.remove(con.getName());
                this.connectionsByUUID.remove(con.getUniqueId());
                this.connectionsByOfflineUUID.remove(con.getPendingConnection().getOfflineId());
            }
        }
        finally {
            this.connectionLock.writeLock().unlock();
        }
    }

    @Override
    public Collection<String> getDisabledCommands() {
        return this.config.getDisabledCommands();
    }

    @Override
    public Collection<ProxiedPlayer> matchPlayer(final String partialName) {
        Preconditions.checkNotNull((Object)partialName, (Object)"partialName");
        final ProxiedPlayer exactMatch = this.getPlayer(partialName);
        if (exactMatch != null) {
            return Collections.singleton(exactMatch);
        }
        return (Collection<ProxiedPlayer>)Sets.newHashSet(Iterables.filter((Iterable)this.getPlayers(), (Predicate)new Predicate<ProxiedPlayer>() {
            public boolean apply(final ProxiedPlayer input) {
                return input != null && input.getName().toLowerCase(Locale.ROOT).startsWith(partialName.toLowerCase(Locale.ROOT));
            }
        }));
    }

    @Override
    public Title createTitle() {
        return new BungeeTitle();
    }

    @Override
    public Configuration getConfig() {
        return this.config;
    }

    @Override
    public PluginManager getPluginManager() {
        return this.pluginManager;
    }

    @Override
    public ReconnectHandler getReconnectHandler() {
        return this.reconnectHandler;
    }

    @Override
    public void setReconnectHandler(final ReconnectHandler reconnectHandler) {
        this.reconnectHandler = reconnectHandler;
    }

    @Override
    public ConfigurationAdapter getConfigurationAdapter() {
        return this.configurationAdapter;
    }

    @Override
    public void setConfigurationAdapter(final ConfigurationAdapter configurationAdapter) {
        this.configurationAdapter = configurationAdapter;
    }

    @Override
    public File getPluginsFolder() {
        return this.pluginsFolder;
    }

    @Override
    public BungeeScheduler getScheduler() {
        return this.scheduler;
    }

    @Override
    public Logger getLogger() {
        return this.logger;
    }

    public ConnectionThrottle getConnectionThrottle() {
        return this.connectionThrottle;
    }
}
